unit MainUnit;

interface
   uses Windows,Messages,Plugin;

type
// Most of OllyDbg windows are the so called tables. A table consists of table
// descriptor (t_table) with embedded sorted data (t_table.sorted, unused in
// custom tables). If data is present, all data elements have the same size and
// begin with a 3-dword t_sorthdr: address, size, type. Data is kept sorted by
// address (in our case this is the index of the bookmark), overlapping is not
// allowed. Our bookmark table consists of elements of type t_bookmark.
  p_bookmark = ^t_bookmark;
  t_bookmark = record
    // Obligatory header, its layout _must_ coincide with t_sorthdr!
    index,                // Bookmark index (0..9)
    size,                 // Size of index, always 1 in our case
    types,                 // Type of entry, TY_xxx
    // Custom data.
    addr:ulong;                 // Address of bookmark
  end;

function Mopenbookmarks(pt:p_table;name:PWChar;index,mode:LongInt):LongInt cdecl;
function Mshowbookmarks(pt:p_table;name:PWChar;index,mode:LongInt):LongInt cdecl;
function Mabout(pt:p_table;name:PWChar;index,mode:LongInt):LongInt cdecl;

function swprintf(Output: PWChar; Format: PWChar): Integer;cdecl;varargs;

function Msetmark(pt:p_table;name:PWChar;index,mode:LongInt):LongInt cdecl;
function Mfollowmark(pt:p_table;name:PWChar;index,mode:LongInt):LongInt cdecl;
function Mdeletemark(pt:p_table;name:PWChar; index,mode:LongInt):LongInt cdecl;

function Bookmarkfunc(pt:p_table;hw,msg:LongWord;wp,lp:LongInt):LongInt stdcall;
function Bookmarkdraw(s:PWChar;mask:PUChar;select:PLongint;pt:p_table;ph:p_drawheader;column:LongInt;cache:Pointer):LongInt cdecl;

var
  bookmark:t_table;             // Bookmark table
  showindisasm:LongInt;         // Option: show bookmarks in Disasm pane


  mainmenu:array[0..3] of t_menu =(
  (name:'Open bookmarks';
   help:'Open Bookmarks window';
   shortcutid:K_NONE;
   menufunc:Mopenbookmarks),
  (name:'Show bookmarks in Disassembler';
  help:'Mark bookmarked command in the Disassembler pane of CPU window';
  shortcutid:K_NONE;
  menufunc:Mshowbookmarks),
  (
  name:'|About';
  help:'About Bookmarks plugin';
  shortcutid:K_NONE;
  menufunc:Mabout),
  ()
  );

  disasmmenu:array[0..18] of t_menu=(
  (name:'';
  help :'Set new bookmark at selection';
  shortcutid:LongInt(KK_DIRECT or KK_CTRL or ord('0'));
  menufunc:Msetmark;
  index:0;
  ),
  (name:'';
  help:'Set new bookmark at selection';
  shortcutid:LongInt(KK_DIRECT or KK_CTRL or ord('1'));
  menufunc:Msetmark;
  index:1;
  ),
  (name:'';
  help:'Set new bookmark at selection';
  shortcutid:LongInt(KK_DIRECT or KK_CTRL or ord('2'));
  menufunc:Msetmark;
  index:2;
  ),
  (name:'';
  help:'Set new bookmark at selection';
  shortcutid:LongInt(KK_DIRECT or KK_CTRL or ord('3'));
  menufunc:Msetmark;
  index:3;
  ),
  (name:'';
  help:'Set new bookmark at selection';
  shortcutid:LongInt(KK_DIRECT or KK_CTRL or ord('4'));
  menufunc:Msetmark;
  index:4;
  ),
  (name:'';
  help:'Set new bookmark at selection';
  shortcutid:LongInt(KK_DIRECT or KK_CTRL or ord('5'));
  menufunc:Msetmark;
  index:5;
  ),
  (name:'|';
  help:'Follow bookmark';
  shortcutid:LongInt(KK_DIRECT or KK_ALT or ord('0'));
  menufunc:Mfollowmark;
  index:0;
  ),
  (name:'';
  help:'Follow bookmark';
  shortcutid:LongInt(KK_DIRECT or KK_ALT or ord('1'));
  menufunc:Mfollowmark;
  index:1;
  ),
  (name:'';
  help:'Follow bookmark';
  shortcutid:LongInt(KK_DIRECT or KK_ALT or ord('2'));
  menufunc:Mfollowmark;
  index:2;
  ),
  (name:'';
  help:'Follow bookmark';
  shortcutid:LongInt(KK_DIRECT or KK_ALT or ord('3'));
  menufunc:Mfollowmark;
  index:3;
  ),
  (name:'';
  help:'Follow bookmark';
  shortcutid:LongInt(KK_DIRECT or KK_ALT or ord('4'));
  menufunc:Mfollowmark;
  index:4;
  ),
  (name:'';
  help:'Follow bookmark';
  shortcutid:LongInt(KK_DIRECT or KK_ALT or ord('5'));
  menufunc:Mfollowmark;
  index:5;
  ),
  // Menu items that delete existing bookmarks.
  (name:'|';
  help:'Delete bookmark';
  shortcutid:LongInt(KK_DIRECT or KK_CTRL or KK_ALT or ord('0'));
  menufunc:Mdeletemark;
  index:0;
  ),
  (name:'';
  help:'Delete bookmark';
  shortcutid:LongInt(KK_DIRECT or KK_CTRL or KK_ALT or ord('1'));
  menufunc:Mdeletemark;
  index:1;
  ),
  (name:'';
  help:'Delete bookmark';
  shortcutid:LongInt(KK_DIRECT or KK_CTRL or KK_ALT or ord('2'));
  menufunc:Mdeletemark;
  index:2;
  ),
  (name:'';
  help:'Delete bookmark';
  shortcutid:LongInt(KK_DIRECT or KK_CTRL or KK_ALT or ord('3'));
  menufunc:Mdeletemark;
  index:3;
  ),
  (name:'';
  help:'Delete bookmark';
  shortcutid:LongInt(KK_DIRECT or KK_CTRL or KK_ALT or ord('4'));
  menufunc:Mdeletemark;
  index:4;
  ),
  (name:'';
  help:'Delete bookmark';
  shortcutid:LongInt(KK_DIRECT or KK_CTRL or KK_ALT or ord('5'));
  menufunc:Mdeletemark;
  index:5;
  ),
  ()
  );

  bookmarkmenu:array[0..0] of t_menu=(
  (name:'|>STANDARD';
  help:'';
  shortcutid:K_NONE)
  );

const
 PLUGIN_NAME      = 'Bookmarks';    // Unique plugin name
 VERSION          = '2.00.00';      // Plugin version

implementation

function memset(src:Pointer;val:LongInt;count:LongWord):Pointer cdecl;
var char_src:PChar;
    i:LongInt;
begin
  char_src := src;
  for i:= count downto 0 do
  begin
    char_src^ := Char(val);
    Inc(char_src);
  end;
  Result := src
end;

function memmove(dest:Pointer;const src:Pointer;count:LongWord):Pointer cdecl;
var dwsrc,
    dwdest:PChar;
begin
  dwdest := dest;
  dwsrc  := src;
  if ((dwdest <= dwsrc) or (dwdest >= (dwsrc + count)))then
  begin
    while(count > 0)do
    begin
      dwdest^ := dwsrc^;
      Inc(dwdest);
      Inc(dwsrc);
      Dec(count);
    end;
  end
  else
  begin
    // overlaping buffers
    dwdest := PChar(dest) + count - 1;
    dwsrc  := PChar(src)  + count - 1;
    while(count > 0)do
    begin
      dwdest^ := dwsrc^;
      Dec(dwdest);
      Dec(dwsrc);
      Dec(count);
    end;  
  end;  
  Result := dest;
end;

function memcpy(dest:Pointer; const src:Pointer;count:LongInt):Pointer cdecl;
var i:LongInt;
begin
  i:= 0;
  while (count > 0)do
  begin
    PUint8(dest)[i] := PUint8(src)[i];
    Inc(i);
    Dec(count);
  end;
  Result := dest;
end;

function wcscpy(dest,src:PWChar):PWChar cdecl;
var i:LongInt;
begin
  i  := 0;
  while src[i] <> #0 do
  begin
    dest[i] := src[i];
    Inc(i);
  end;
  dest[i] := #0;
  Result := dest;
end;

function wcscmp(cs,ct:PWChar):LongInt cdecl;
begin
  Result := 0;
  while (cs^ = ct^)do
  begin
    if (cs^ = #0)then Exit;
    Inc(ct);
    Inc(cs);
  end;
  Result := Byte(cs^) - Byte(ct^);
end;

function Bookmarksortfunc(const sh1,sh2:p_sorthdr;const n:LongInt):LongInt cdecl;
var bm1,bm2:p_bookmark;
    i:LongInt;
begin
  bm1 :=p_bookmark(sh1);
  bm2 :=p_bookmark(sh2);
  i := 0;
  if (n=1) then
  begin
    if (bm1.addr<bm2.addr)then i:=-1
    else if (bm1.addr>bm2.addr)then i:=1;
  end;
  if (i=0)then
  begin
    if (bm1.index<bm2.index)then i:=-1
    else if (bm1.index>bm2.index)then i:=1;
  end;
  Result := i;
end;

procedure Bookmarkdestfunc(sh:p_sorthdr)cdecl;begin end;

// ODBG_Pluginquery() is a "must" for valid OllyDbg plugin. First it must check
// whether given OllyDbg version is correctly supported, and return 0 if not.
// Then it should make one-time initializations and allocate resources. On
// error, it must clean up and return 0. On success, if should fill plugin name
// and plugin version (as UNICODE strings) and return version of expected
// plugin interface. If OllyDbg decides that this plugin is not compatible, it
// will call ODBG2_Plugindestroy() and unload plugin. Plugin name identifies it
// in the Plugins menu. This name is max. 31 alphanumerical UNICODE characters
// or spaces + terminating L'\0' long. To keep life easy for users, this name
// should be descriptive and correlate with the name of DLL. This function
// replaces ODBG_Plugindata() and ODBG_Plugininit() from the version 1.xx.
function Pluginquery(ollydbgversion:LongInt;pluginname,pluginversion:PWChar):LongInt cdecl;
var restore:LongInt;
begin
  Result := 0;
  if (ollydbgversion < PLUGIN_VERSION)then Exit;
  // Initialize bookmark storage. Data contains no resources, so destructor is
  // not necessary. (Destructor is called each time data item is removed from
  // the sorted data). I add it pro forma, just to remind you of this option.
  Createsorteddata(@(bookmark.sorted), // Descriptor of sorted data
    sizeof(t_bookmark),                // Size of single data item
    10,                                // Initial number of allocated items
    @Bookmarksortfunc,      // Sorting function
    @Bookmarkdestfunc,      // Data destructor
    0);                                // Simple data, no special options
  // Initialize bookmark table. OllyDbg uses table name to save the appearance
  // to the main initialization file. Keep this name unique, or else.

  StrcopyW(bookmark.name,SHORTNAME,PLUGIN_NAME);
  bookmark.mode:=TABLE_SAVEALL;         // Save complete appearance
  bookmark.bar.visible:=1;              // By default, bar is visible
  bookmark.bar.name[0]:='Bookmark';
  bookmark.bar.expl[0]:='Bookmark index';
  bookmark.bar.mode[0]:=BAR_SORT;
  bookmark.bar.defdx[0]:=9;
  bookmark.bar.name[1]:='Address';
  bookmark.bar.expl[1]:='Bookmark address';
  bookmark.bar.mode[1]:=BAR_SORT;
  bookmark.bar.defdx[1]:=9;
  bookmark.bar.name[2]:='Disassembly';
  bookmark.bar.expl[2]:='Command at the bookmark address';
  bookmark.bar.mode[2]:=BAR_FLAT;
  bookmark.bar.defdx[2]:=24;
  bookmark.bar.name[3]:='Comments';
  bookmark.bar.expl[3]:='Comments';
  bookmark.bar.mode[3]:=BAR_FLAT;
  bookmark.bar.defdx[3]:=256;
  bookmark.bar.nbar:=4;
  bookmark.tabfunc:=@Bookmarkfunc;
  bookmark.custommode:=0;
  bookmark.customdata:=nil;
  bookmark.updatefunc:=nil;
  bookmark.drawfunc:=@Bookmarkdraw;
  bookmark.tableselfunc:=nil;
  bookmark.menu:=@bookmarkmenu;
  // Get initialization data.
  showindisasm:=0;                      // Default value
  Getfromini(nil,PLUGIN_NAME,'Show bookmarks in Disassembler','%i',
    @showindisasm);
  // OllyDbg saves positions of plugin windows with attribute TABLE_SAVEPOS to
  // the .ini file but does not automatically restore them. Let us add this
  // functionality here. To conform to OllyDbg norms, window is restored only
  // if corresponding option is enabled.
  if (restorewinpos<>0) then
  begin
    restore:=0;                         // Default
    Getfromini(nil,PLUGIN_NAME,'Restore window','%i',@restore);
    if (restore <> 0)then
    Createtablewindow(@bookmark,0,bookmark.bar.nbar,0,'ICO_P',PLUGIN_NAME);
  end;  
  // Report name and version to OllyDbg.
  wcscpy(pluginname,PLUGIN_NAME);       // Name of plugin
  wcscpy(pluginversion,VERSION);       // Version of plugin
  Result := PLUGIN_VERSION;
end;

function Pluginmenu(types:PWChar):p_menu cdecl;
begin
  Result := nil;
  if (wcscmp(types,PWM_MAIN)=0)then
  begin
    // Main menu.
    Result := @mainmenu;
  end
  else
  if (wcscmp(types,PWM_DISASM)=0)then
  begin
    // Disassembler pane of CPU window.
    Result := @disasmmenu;
  end;  
end;

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////// DUMP WINDOW HOOK ///////////////////////////////

// Dump windows display contents of memory or file as bytes, characters,
// integers, floats or disassembled commands. Plugins have the option to modify
// the contents of the dump windows. If ODBG2_Plugindump() is present and some
// dump window is being redrawn, this function is called first with column=
// DF_FILLCACHE, addr set to the address of the first visible element in the
// dump window and n to the estimated total size of the data displayed in the
// window (n may be significantly higher than real data size for disassembly).
// If plugin returns 0, there are no elements that will be modified by plugin
// and it will receive no other calls. If necessary, plugin may cache some data
// necessary later. OllyDbg guarantees that there are no calls to
// ODBG2_Plugindump() from other dump windows till the final call with
// DF_FREECACHE.
// When OllyDbg draws table, there is one call for each table cell (line/column
// pair). Parameters s (UNICODE), mask (DRAW_xxx) and select (extended DRAW_xxx
// set) contain description of the generated contents of length n. Plugin may
// modify it and return corrected length, or just return the original length.
// When table is completed, ODBG2_Plugindump() receives final call with
// column=DF_FREECACHE. This is the time to free resources allocated on
// DF_FILLCACHE. Returned value is ignored.
// Use this feature only if absolutely necessary, because it may strongly
// impair the responsiveness of the OllyDbg. Always make it switchable with
// default set to OFF!
function Plugindump(pd:p_dump;s:PWChar;mask:PUChar;n:LongInt;
                    select:PLongint;addr:ulong;column:LongInt):LongInt cdecl ;
var i,j:LongInt;
    w:array[0..TEXTLEN-1]of wchar_t;
    pmark:p_bookmark;
begin
  Result := 0;
  if (column=DF_FILLCACHE)then
  begin
    // Check whether this feature is requested.
    if (showindisasm=0)then
    Exit;  // Turned off
    // Check whether it's Disassembler pane of the CPU window.
    if (pd=nil )or( (pd.menutype and DMT_CPUMASK)<>DMT_CPUDASM)then
    Exit;  // Not a Disassembler
    // Just for the sake, assure that bookmarks apply: not a file dump, not a
    // backup display
    if (pd.filecopy<>nil )or(
       (pd.dumptype and DU_TYPEMASK)<>DU_DISASM )or(
       (pd.dumptype and DU_BACKUP)<>0)
    then
    Exit;  // Invalid dump type
    // Check that there are bookmarks in the supplied memory range.
    for I := 0 to bookmark.sorted.n - 1 do
    begin
      pmark:=p_bookmark(Getsortedbyindex(@(bookmark.sorted),i));
      if (pmark=nil)then  continue;                      // Must not happen!

      if (pmark.addr>=addr )and( pmark.addr<addr+ LongWord(n))then
      begin
        Result := 1;       // Bookmark to display
        Exit;
      end;
    end;
    Exit;        // No bookmarks to display
  end
  else
  if (column=2)then
  begin
    // Check whether there is a bookmark. Note that there may be several marks
    // on the same address!
    w[0]:='['; j:=1;
    for I := 0 to bookmark.sorted.n - 1 do
    begin
      pmark:=p_bookmark(Getsortedbyindex(@(bookmark.sorted),i));
      if (pmark=nil)then  continue;         // Must not happen!
      if (pmark.addr<>addr)then  continue;
      if (j>1)then
      begin
        w[j]:=',';                   // Comma between bookmarks
        Inc(j);
      end;
      Inc(j,swprintf(w+j,'%i',pmark.index));
    end;
    if (j=1)then
    begin
      Result := n;        // No bookmarks on address
      Exit;
    end;
    Inc(j,StrcopyW(w+j,TEXTLEN-j,'] '));
    // Disassebly column is always drawn by mask (DRAW_MASK bit set). But this
    // is a sample plugin and I show here what to do if this is not the case.
    if ((select^ and DRAW_MASK)=0) then
    begin
      memset(mask,DRAW_NORMAL,n);
      select^:= select^ or DRAW_MASK;
    end;
    // Skip graphical symbols (loop brackets).
    i := 0;
    while i<n do
    begin   
      if ((PUint8(mask)[i] and DRAW_GRAPH)=0)then break;
      Inc(i);
    end;
    // Insert text.
    Inc(n,j);
    if (n>TEXTLEN)then
    n:=TEXTLEN;
    if (n>i+j) then
    begin
      memmove(s+i+j,s+i,(n-i-j)*sizeof(wchar_t));
      memmove(PChar(mask)+i+j,PChar(mask)+i,n-i-j);
    end;
    memcpy(s+i,@w,j*sizeof(wchar_t));
    memset(PChar(mask)+i,DRAW_EIP,j-1);
    PUint8(mask)[i+j-1]:=DRAW_NORMAL;
  end
  else
  if (column=DF_FREECACHE)then
  begin
    // We have allocated no resources, so we have nothing to do here.
  end;
  Result := n;
end;

procedure Pluginmainloop(debugevent:PDebugEvent)cdecl;
begin

end;

// Drawing function of Bookmarks window.
function Bookmarkdraw;
var
  length:ULong;
  cmd:array[0..MAXCMDSIZE-1]of Byte;
  m,n:LongInt;          // Number of symbols in the string
  pmark:p_bookmark;
  pasm:p_disasm;
begin
  n:=0;
  Result := 0;
  // For simple tables, t_drawheader is the pointer to the data element. It
  // can't be NULL, except in DF_CACHESIZE, DF_FILLCACHE and DF_FREECACHE.
  pmark:=p_bookmark(ph);
  // Our cache is just a t_disasm. It is not NULL on the same conditions.
  pasm:=p_disasm(cache);
  case column of
    DF_CACHESIZE:    // Request for draw cache size
    begin
      // Columns 3 and 4 (disassembly and comment) both require calls to
      // Disasm(). To accelerate processing, I call disassembler once per line
      // and cache data between the calls. Here I inform the drawing routine
      // how large the cache must be.
      Result:= sizeof(t_disasm);
      Exit;
    end;  
    DF_FILLCACHE: // Request to fill draw cache
    begin
      // We don't need to initialize cache when drawing begins. Note that cache
      // is initially zeroed.
      Exit;
    end;
    DF_FREECACHE:                 // Request to free cached resources
    begin
      // We don't need to free cached resources when drawing ends.
      Exit;
    end;
    DF_NEWROW:                    // Request to start new row in window
    begin
      // New row starts. Let us disassemble the command at the pointed address.
      // I assume that bookmarks can't be set on data. First of all, we need to
      // read the contents of memory. Length of 80x86 commands is limited to
      // MAXCMDSIZE bytes.
      length:=Readmemory(@cmd,pmark.addr,sizeof(cmd),MM_SILENT or MM_PARTIAL);
      if (length=0) then
      begin
        // Memory is not readable.
        StrcopyW(pasm.result,TEXTLEN,'???');
        StrcopyW(pasm.comment,TEXTLEN,'');
      end
      else
      begin
        Disasm(@cmd,length,pmark.addr,PUChar(Finddecode(pmark.addr,nil)),pasm,
        DA_TEXT or DA_OPCOMM or DA_MEMORY,nil,nil);
      end;
      Exit;
    end;
    0:                            // 0-based index
    begin
      n:=StrcopyW(s,TEXTLEN,'Alt+');
      memset(mask,DRAW_GRAY,n);
      m:=swprintf(s+n,'%i',pmark.index);
      memset(PChar(mask)+n,DRAW_NORMAL,m);
      Inc(n,m);
      select^:=select^ or DRAW_MASK;
    end;
    1:                            // Address of the bookmark
    begin
      n:=Simpleaddress(s,pmark.addr,mask,select);
    end;
    2:                            // Disassembled command
    begin
      n:=StrcopyW(s,TEXTLEN,pasm.result);
    end;
    3:                            // Comment
    begin
      // User-defined comment has highest priority.
      n:=FindnameW(pmark.addr,NM_COMMENT,s,TEXTLEN);
      // Comment created by Disasm() is on the second place.
      if (n=0)then
        n:=StrcopyW(s,TEXTLEN,pasm.comment);
      // Analyser comment follows.
      if (n=0)then
        n:=Getanalysercomment(nil,pmark.addr,s,TEXTLEN);
      // Procedure comments have the lowest priority.
      if (n=0)then
        n:=Commentaddress(pmark.addr,COMM_MARK or COMM_PROC,s,TEXTLEN);
    end;    
  end;
  Result := n;
end;

function Bookmarkfunc;
var pmark:p_bookmark;
begin
  Result := 0;
  case msg of
    WM_USER_DBLCLK:               // Doubleclick
    begin
      // Get selection.
      pmark:=p_bookmark(Getsortedbyselection(
        @(pt.sorted),pt.sorted.selected));
      // Follow address in CPU Disassembler pane. Actual address is added to
      // the history, so that user can easily return back to it.
      if (pmark <> nil)then Setcpu(0,pmark.addr,0,0,0,
        CPU_ASMHIST or CPU_ASMCENTER or CPU_ASMFOCUS);
      Result := 1;
    end;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
////////////////// PLUGIN MENUS EMBEDDED INTO OLLYDBG WINDOWS //////////////////

// Menu processing functions are called twice. First time (mode=MENU_VERIFY)
// OllyDbg asks to verify whether corresponding menu item applies or not. If
// necessary, menu function may change menu text (parameter name, up to TEXTLEN
// UNICODE characters). It must return one of the following codes:
//
//   MENU_ABSENT:     menu item does not apply and should not be present in
//                    the menu;
//   MENU_NORMAL:     menu item appears in the menu;
//   MENU_CHECKED:    menu item appears in the menu and has attached checkmark;
//   MENU_CHKPARENT:  menu item appears in the menu and has attached checkmark.
//                    Additionally, attaches checkmark to the parent item in
//                    menu on the previous level. This feature does not work in
//                    the main menu;
//   MENU_SHORTCUT:   menu item does not appear in the menu but is active and
//                    participates in the search for keyboard shortcut;
//   MENU_GRAYED:     item is present in the menu but disabled. This style is
//                    not compatible with OllyDbg's look-and-feel, use it only
//                    if absolutely necessary due to the menu logic.
//
// When menu item is selected (mouseclick or keyboard shortcut), menu function
// is called for the second time (mode=MENU_EXECUTE, name is undefined). At
// this moment, all threads of the debugged application are suspended. Function
// must make all necessary actions and return one of the following codes:
//
//   MENU_REDRAW:     this action has global impact, all OllyDbg windows must
//                    be updated. OllyDbg broadcasts WM_USER_CHGALL;
//   MENU_NOREDRAW:   no redrawing is necessary.
//
// If processing is lengthy and application should continue execution, use
// Resumeallthreads() at entry to the MENU_EXECUTE block and Suspendallthreads()
// on exit. Note that MENU_ABSENT and MENU_NOREDRAW are interchangeable.
//
// Parameter index allows to use single menu function with several similar menu
// items.
//
// Note that OllyDbg uses menu structuress to process keyboard shortcuts. It is
// done automatically, without the need to pay additional attention.

// Menu function of main menu, opens or brings to top list of bookmarks.

function Mopenbookmarks;
begin
  Result := MENU_ABSENT;
  case mode of
    MENU_VERIFY: Result := MENU_NORMAL;   // Always available
    MENU_EXECUTE:
    begin
      if (bookmark.hw=0)then
      begin
        Createtablewindow(@bookmark,0,bookmark.bar.nbar,0,'ICO_P',PLUGIN_NAME);
      end
      else
      Activatetablewindow(@bookmark);
      Result := MENU_NOREDRAW;
    end;
  end;
end;

function Mshowbookmarks ;
begin
  Result := MENU_ABSENT;
  case mode of
    MENU_VERIFY:
    begin
      if showindisasm <> 0 then  Result := MENU_CHECKED
      else Result := MENU_NORMAL;
    end;  
    MENU_EXECUTE:
    begin
      showindisasm:= not showindisasm;
      Writetoini(nil,PLUGIN_NAME,'Show bookmarks in Disassembler','%i',
      showindisasm);
      Result := MENU_REDRAW;
    end;
  end;
end;

function Mabout ;
var n:LongInt;
    s:array[0..TEXTLEN-1]of WideChar;
begin
  Result := MENU_ABSENT;
  case mode of
    MENU_VERIFY: Result := MENU_NORMAL;
    MENU_EXECUTE:
    begin
      // Debuggee should continue execution while message box is displayed.
      Resumeallthreads;
      // In this case, swprintf() would be as good as a sequence of StrcopyW(),
      // but secure copy makes buffer overflow impossible.
      n:=StrcopyW(s,TEXTLEN,'Bookmark plugin v');
      Inc(n,StrcopyW(s+n,TEXTLEN-n,VERSION));
      // COPYRIGHT POLICY: This bookmark plugin is an open-source freeware. It's
      // just a sample. The copyright below is also just a sample and applies to
      // the unmodified sample code only. Replace or remove copyright message if
      // you make ANY changes to this code!
      StrcopyW(s+n,TEXTLEN-n,#10'Copyright (C) 2001-2011 Oleh Yuschuk');
      MessageBoxW(hwollymain,s,
        'Bookmark plugin',MB_OK or MB_ICONINFORMATION);
      // Suspendallthreads() and Resumeallthreads() must be paired, even if they
      // are called in inverse order!
      Suspendallthreads;
      Result := MENU_NOREDRAW;
    end;  
  end;
end;

// Menu function of Disassembler pane that creates new bookmark.
function Msetmark;
var pd:p_dump;
    pmark:p_bookmark;
    bmk:t_bookmark;
begin
  Result := MENU_ABSENT;
  if (pt=nil )or( pt.customdata=nil)then  Exit;

  pd:=p_dump(pt.customdata);
  if (mode=MENU_VERIFY) then
  begin
    // Check whether selection is defined.
    if (pd.sel0>=pd.sel1)then Exit;
    // Check whether bookmark with given index already exists.
    pmark:=p_bookmark(Findsorteddata(@(bookmark.sorted),index,0));
    if (pmark <> nil)then  Exit;
    swprintf(name,'Set bookmark %i',index);
    Result := MENU_NORMAL;
  end
  else
  if (mode=MENU_EXECUTE) then
  begin
    bmk.index:=index;
    bmk.size:=1;
    bmk.types:=0;
    bmk.addr:=pd.sel0;
    Addsorteddata(@(bookmark.sorted),@bmk);
    Pluginmodulechanged(pd.sel0);
    Result := MENU_REDRAW;
  end;  
end;

// Menu function of Disassembler pane that follows existing bookmark.
function Mfollowmark;
var pd:p_dump;
    pmark:p_bookmark;
begin
  Result := MENU_ABSENT;
  // Get dump descriptor.
  if (pt=nil )or( pt.customdata=nil)then  Exit;
  pd:=p_dump(pt.customdata);
  // Check whether bookmark is defined.
  pmark:=p_bookmark(Findsorteddata(@(bookmark.sorted),index,0));
  if (pmark=nil)then  Exit;
  if (mode=MENU_VERIFY) then
  begin
    // If we are already at bookmark, precede its name with checkmark.
    if (pd.sel0<pd.sel1 )and( pmark.addr=pd.sel0)then
    begin
      swprintf(name,'Bookmark %i',index);
      Result := MENU_CHECKED;
      Exit;
    end;
    swprintf(name,'Follow bookmark %i',index);
    Result := MENU_NORMAL;
  end
  else
  if (mode=MENU_EXECUTE)then
  begin
    Setcpu(0,pmark.addr,0,0,0,
      CPU_ASMHIST or CPU_ASMCENTER or CPU_ASMFOCUS);
    Result := MENU_REDRAW;
  end;     
end;

// Menu function of Disassembler pane that deletes existing bookmark.
function Mdeletemark;
var pmark:p_bookmark;
    pd:p_dump;
begin
  Result := MENU_ABSENT;
  // Get dump descriptor.
  if (pt=nil )or( pt.customdata=nil)then Exit;
  pd:=p_dump(pt.customdata);
  case mode of
    MENU_VERIFY:
    begin
      // Check whether bookmark is defined.
      pmark:=p_bookmark(Findsorteddata(@(bookmark.sorted),index,0));
      if (pmark=nil)then Exit;
      swprintf(name,'Delete bookmark %i',index);
      // If we are already at bookmark, precede its name with checkmark.
      if (pd.sel0<pd.sel1 )and( pmark.addr=pd.sel0)then
      begin
        Result := MENU_CHECKED;
        Exit;
      end;
      Result := MENU_NORMAL;
    end;  
    MENU_EXECUTE:
    begin
      Deletesorteddata(@(bookmark.sorted),index,0);
      Result := MENU_REDRAW;
    end;  
  end;    
end;

procedure Pluginanalyse(pmod:P_module)cdecl;begin end;

procedure Pluginexception(preg:p_reg)cdecl;
begin

end;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////// .UDD FILES //////////////////////////////////

// For each plugin, OllyDbg automatically creates file-unique block of .udd
// record types. They are indexed by plugin names. If there are two plugins
// with the same name, restored records may be sent to the wrong plugin.
// Therefore it is very important that plugin names are unique.

const BOOKMARKTAG    =$0001;          // Tag of bookmark record

type
  p_uddmark = ^t_uddmark;
  t_uddmark = record             // Bookmark in .udd file
    index:LongInt;               // Bookmark index
    offset:ulong;                // Offset from module base
  end;

// Time to save data to .udd file! This is done by calling Pluginsaverecord()
// for each data item that must be saved. Global, process-oriented data must
// be saved in main .udd file (named by .exe); module-relevant data must be
// saved in module-related files. Don't forget to save all addresses relative
// to module's base, so that data will be restored correctly even when module
// is relocated.
procedure Pluginsaveudd(psave:p_uddsave;pmod:p_module;ismainmodule:LongInt)cdecl;
var pmark:p_bookmark;
    i:LongInt;
    um:t_uddmark;
begin
  for I := 0 to bookmark.sorted.n - 1 do
  begin
    pmark:=p_bookmark(Getsortedbyindex(@(bookmark.sorted),i));
    if (pmark=nil)then continue;
    if (pmark.addr>=pmod.base )and( pmark.addr<pmod.base+pmod.size) then
    begin
      // Bookmarks are saved with corresponding modules. If bookmark was set
      // outside of any module, it will be lost.
      um.index:=pmark.index;
      um.offset:=pmark.addr-pmod.base;
      Pluginsaverecord(psave,BOOKMARKTAG,sizeof(um),@um);
    end;
  end;
end;

// OllyDbg restores data from .udd file and has encountered record belonging
// to our plugin. Note that module descriptor pointed to by pmod can be
// incomplete, i.e. does not necessarily contain all informations, especially
// that from .udd file.
procedure Pluginuddrecord(pmod:p_module;ismainmodule:LongInt;tag, size:ulong;data:Pointer)cdecl;
var
   bmk:t_bookmark;
   pum:p_uddmark;
begin
  if (pmod=nil )or( data=nil)then Exit;  // Must not happen!
  if (size<>sizeof(t_uddmark))then Exit; // Invalid record size
  if (tag<>BOOKMARKTAG)then  Exit;       // Invalid tag
  pum:=p_uddmark(data);
  if (pum.index<0 )or( pum.index>9)then  Exit;   // Corrupted data
  bmk.index:=pum.index;
  bmk.size:=1;
  bmk.types:=0;
  bmk.addr:=pmod.base+pum.offset;
  Addsorteddata(@(bookmark.sorted),@bmk);
end;

// Function is called when user opens new or restarts current application.
// Plugin should reset internal variables and data structures to initial state.
procedure Pluginreset;cdecl;
begin
  Deletesorteddatarange(@(bookmark.sorted),0,$FFFFFFFF);
end;

// OllyDbg calls this optional function when user wants to terminate OllyDbg.
// All MDI windows created by plugins still exist. Function must return 0 if
// it is safe to terminate. Any non-zero return will stop closing sequence. Do
// not misuse this possibility! Always inform user about the reasons why
// termination is not good and ask for his decision! Attention, don't make any
// unrecoverable actions for the case that some other plugin will decide that
// OllyDbg should continue running.
function Pluginclose:LongInt cdecl;
begin
  // For automatical restoring of open windows, mark in .ini file whether
  // Bookmarks window is still open.
  Writetoini(nil,PLUGIN_NAME,'Restore window','%i',bookmark.hw<>0);
  Result := 0;
end;

// OllyDbg calls this optional function once on exit. At this moment, all MDI
// windows created by plugin are already destroyed (and received WM_DESTROY
// messages). Function must free all internally allocated resources, like
// window classes, files, memory etc.
procedure Plugindestroy;cdecl;
begin
  Destroysorteddata(@(bookmark.sorted));
end;

function swprintf; external user32 name 'wsprintfW';

exports
   Plugindestroy       name '_ODBG2_Plugindestroy',
   Pluginquery         name '_ODBG2_Pluginquery',
   Pluginanalyse       name '_ODBG2_Pluginanalyse',
   Pluginmenu          name '_ODBG2_Pluginmenu',
   Pluginmainloop      name '_ODBG2_Pluginmainloop',
   Pluginexception     name '_ODBG2_Pluginexception',
   Pluginsaveudd       name '_ODBG2_Pluginsaveudd',
   Pluginuddrecord     name '_ODBG2_Pluginuddrecord',
   Pluginreset         name '_ODBG2_Pluginreset',
   Pluginclose         name '_ODBG2_Pluginclose',
   Plugindump          name '_ODBG2_Plugindump';

end.
